#include <irctrans.h>

/*
finds the first default session (should only be one) for a jid.
*/
session irc_find_default_session(jid j)
/*
returns:
 NULL if no default session was found
 a session pointer to the default session for that jid
input:
 jid j: the jid you want to find a default session for
*/
{
    session s;
    if(j==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    for(s=sessions;s!=NULL;s=s->next)
    {
        if(jid_cmp(j,s->j)==0)
            if(s->default_session==1)
            {
                pth_mutex_release(&s__m);
                return s;
            }
    }
    pth_mutex_release(&s__m);
    return NULL;    
}  

/*
does a resource insensitive compare, NULL != NULL
*/
int jid_cmp_nores(jid j,jid r)
/*
returns:
 0 if a match is found
 -1 if either term is null
 any other return val that strcmp returns
input:
 jid j: the first jid to compare, this pool is added to in this function
 jid r: the other jid to compare, this pool is untouched
*/
{
    jid j2,r2;
    if(j==NULL||r==NULL) return -1;
    j2=jid_new(j->p,jid_full(j));
    r2=jid_new(j->p,jid_full(r));
    jid_set(j2,NULL,JID_RESOURCE);
    jid_set(r2,NULL,JID_RESOURCE);
    return jid_cmp(j2,r2);
}

/*
handles an unsubscribed packet to register.irc.foo.dom
if this is for the transport item, we should be logging them
out of the default session.
*/
void unsubscribed_user(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the jpacket from etherx
*/
{ 
    session s;
    if(jp==NULL) return;
    if(jp->to->user!=NULL)
    { /* unsubscribed to a user.. who cares?! */
        xmlnode_free(jp->x);
        return;
    }
    /* find default session and set to 0 */
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_find_default_session(jp->from);
    
    if(s!=NULL&&s->rc!=0)
    { /* a default session was found, kill it */
        s->default_session=0;
        s->rc--;
        if(s->rc==0)
        {
            session_unlink(s);    
            tstream_close(s->stream);    
        }
        jutil_tofrom(jp->x);
        xmlnode_put_attrib(jp->x,"type","unavailable");
        jpacket_reset(jp);
        ehandler_send(ehandler_reg,xmlnode_dup(jp->x),jp->to->server);
        while(s->notify_list!=NULL)
        {
            spool sp=spool_new(jp->p);
            spooler(sp,s->notify_list->user,"@",reg_host,sp);
            xmlnode_put_attrib(jp->x,"from",spool_print(sp));
            ehandler_send(ehandler_reg,xmlnode_dup(jp->x),jp->to->server);
            s->notify_list=s->notify_list->next;
        }
    }
    pth_mutex_release(&s__m);
    xmlnode_free(jp->x);
}

/*
handles a jabber subscription to an IRC user
add it to our notify list, (so ISON can query the presence)
*/
void subscribe_user(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the incoming subscription packet from etherx
*/
{
    if(jp==NULL)return;
    jutil_tofrom(jp->x);
    jpacket_reset(jp);
    irc_add_notify(irc_find_default_session(jp->from),jp->to->user);
    xmlnode_put_attrib(jp->x,"type","subscribed");
    ehandler_send(ehandler_irc,xmlnode_dup(jp->x),jp->to->server);
    xmlnode_put_attrib(jp->x,"type","subscribe");
    ehandler_send(ehandler_irc,jp->x,jp->to->server);
}

/*
takes an available presence packet, and logs the user into IRC in a default session
*/
void irc_auth_user(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the incoming presence from etherx
*/
{
    xmlnode xres;
    session s;
    xmlnode presence;
    spool sp=spool_new(jp->p);
    char *user,*server,*password;
    jpacket logon_jp;
    if(jp!=NULL&&jp->to->user!=NULL)
    {
        irc_parse_reg_available(jp);
        return;
    }
    if(jp==NULL||jp->to->resource==NULL||(xres=jid_xres(jp->to))==NULL)
    {
        xmlnode_free(jp->x);
        return;
    }
    user=xmlnode_get_attrib(xres,"user");
    server=xmlnode_get_attrib(xres,"server");
    password=xmlnode_get_attrib(xres,"pass");
    if(user==NULL||server==NULL)
    {
        xmlnode_free(jp->x);
        return;
    }
    /* 
    need to convert to NULL%server@irc.foo.dom/user 
    becuase session_create will kill the jpacket,
    and also expects the to/from in a certain format
    */
    spooler(sp,"NULL%",server,"@",irc_host,"/",user,sp);
    presence=xmlnode_new_tag("presence");
    xmlnode_put_attrib(presence,"to",spool_print(sp));
    xmlnode_put_attrib(presence,"from",jid_full(jp->from));
    xmlnode_put_attrib(presence,"pass",password);
    logon_jp=jpacket_new(presence);
    if(logon_jp->to==NULL)
    {
        xmlnode_free(logon_jp->x);
        xmlnode_free(jp->x);
        return;
    }
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_session_find_by_jid(jp->from,NULL);
    if(s==NULL&&(s=irc_find_default_session(jp->from))==NULL)
    {
        irc_session_create(logon_jp,1,password);
    }
    else if(s!=NULL&&s->default_session!=1)
    {
        xmlnode_free(logon_jp->x);
        s->default_session=1;
        s->rc++;
    }
    pth_mutex_release(&s__m);
    jutil_tofrom(jp->x);
    jpacket_reset(jp);
    ehandler_send(ehandler_reg,jp->x,jp->to->server);
}

/*
gets an available packet send to an IRC user,
so that it can be added to the notify list
*/
int irc_parse_reg_available(jpacket jp)
/*
returns:
 0 always
input:
 jpacket jp: the incoming packet from etherx
*/
{
    /* available for a user */
    pth_yield(NULL); /* give the session a chance to start */
    irc_add_notify(irc_find_default_session(jp->from),jp->to->user);
    xmlnode_free(jp->x);
    return 0;
}

/*
handles an unavailable packet from etherx..
this can be either to a user or to the session
*/
int irc_parse_reg_unavailable(jpacket jp)
/*
returns:
 1 on error
 0 on normal exit
*/
{
    session s;
    if(jp==NULL) return 1;
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_find_default_session(jp->from);
    if(s!=NULL&&jp->to->user!=NULL)
    { /* unavail to a user, del it from the notify list */
        irc_del_notify(s,jp->to->user);
        pth_mutex_release(&s__m);
        jutil_tofrom(jp->x);
        jpacket_reset(jp);
        ehandler_send(ehandler_reg,jp->x,jp->to->server);
        return 0;
    }
    log_debug(ZONE, "[parser] Unavailable for Default session of (%s)", jid_full(jp->from) );
    if(s!=NULL&&s->rc!=0)
    {
        s->rc--;
        s->default_session=0;
    }
    jutil_tofrom(jp->x);
    jpacket_reset(jp);
    if(s==NULL)
    {
        pth_mutex_release(&s__m);
        ehandler_send(ehandler_reg,jp->x,jp->to->server);
        return 1;
    }
    else
        ehandler_send(ehandler_reg,xmlnode_dup(jp->x),jp->to->server);
    /* send an unavail for each user as well */
    while(s->notify_list!=NULL)
    {
        spool sp=spool_new(jp->p);
        spooler(sp,s->notify_list->user,"@",reg_host,sp);
        xmlnode_put_attrib(jp->x,"from",spool_print(sp));
        jpacket_reset(jp);
        ehandler_send(ehandler_reg,xmlnode_dup(jp->x),jp->to->server);
        s->notify_list=s->notify_list->next;
    }
    if(s->rc==0)
    {
        session_unlink(s);    
        tstream_close(s->stream);    
    }
    pth_mutex_release(&s__m);
    xmlnode_free(jp->x);
    return 0;
}

/*
handles a message from jabber to an IRC user,
using the register.irc.foo.dom namespace
*/
void irc_parse_reg_message(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the incoming jpacket from etherx
*/
{
    session s;
    spool sp;
    char *sendmsg;
    if(jp==NULL)return;
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_find_default_session(jp->from);
    sp=spool_new(jp->p);
    sendmsg=UTF8_to_str(jp->p,xmlnode_get_data(xmlnode_get_firstchild(xmlnode_get_tag(jp->x,"body"))));
    if(s==NULL||s->rc==0||sendmsg==NULL) 
    {
        pth_mutex_release(&s__m);
        xmlnode_free(jp->x);
        return;
    }    
    if(strlen(sendmsg)>450)
        sendmsg[450]='\0';
    if(strncasecmp(sendmsg,"/me ",4)==0)
    {
        /* this should be an action */
        jutil_tofrom(jp->x);
        sendmsg+=4;
        spooler(sp,"PRIVMSG ",jp->to->user," :\001ACTION ",sendmsg,"\001\n",sp);
        tstream_write(s->stream,spool_print(sp),-1);
        xmlnode_free(jp->x);
    }
    else /* standard message */
    {
        spooler(sp,"PRIVMSG ",jp->to->user," :",sendmsg,"\n",sp);
        tstream_write(s->stream,spool_print(sp),-1);
        xmlnode_free(jp->x);
    }
    pth_mutex_release(&s__m);
}    

/*
respond to an <iq type='get'><query xmlns='jabber:iq:register'/></iq> packet
*/
void irc_parse_iqget(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the incoming jpacket from etherx
*/
{
    xmlnode query;
    session s;
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_find_default_session(jp->from);
    jutil_iqresult(jp->x);
    query=xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(query,"xmlns","jabber:iq:register");
    xmlnode_insert_cdata(xmlnode_insert_tag(query,"instructions"),"Insert your nick in the Username field, The server in the Misc field, and any nickserv password in the password field",-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(query,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);
    if(s==NULL)
    {
        xmlnode_insert_tag(query,"username");
        xmlnode_insert_tag(query,"misc");
        xmlnode_insert_tag(query,"password");
    }
    else
    {
        xmlnode_insert_cdata(xmlnode_insert_tag(query,"username"),s->nick,-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(query,"misc"),s->server,-1);
        xmlnode_insert_tag(query,"password");
    }
    pth_mutex_release(&s__m);
    jpacket_reset(jp);
    ehandler_send(ehandler_irc,jp->x,jp->to->server);
}

/*
take a jabber:iq:register set, and start a default session for the user.
*/
void irc_parse_iqset(jpacket jp)
/*
returns:
 nothing
input:
 jpacket jp: the incoming jpacket from etherx
*/
{
    session s;
    spool sp;
    xmlnode logon;
    xmlnode presence,query;
    char *nick,*server,*pass;
    if(jp==NULL)return;

    sp=spool_new(jp->p);
    /* verify the packet */
    query=xmlnode_get_tag(jp->x,"query");
    if(query==NULL)
    {
        xmlnode_free(jp->x);
        return;
    }
    nick=xmlnode_get_data(xmlnode_get_firstchild(xmlnode_get_tag(query,"username")));
    server=xmlnode_get_data(xmlnode_get_firstchild(xmlnode_get_tag(query,"misc")));
    pass=xmlnode_get_data(xmlnode_get_firstchild(xmlnode_get_tag(query,"password")));
    if(nick==NULL||server==NULL)
    {
        xmlnode err=xmlnode_new_tag("message"),error;
        xmlnode_put_attrib(err,"type","error");
        xmlnode_put_attrib(err,"to",jid_full(jp->from));
        xmlnode_put_attrib(err,"from",jid_full(jp->to));
        error=xmlnode_insert_tag(err,"error");
        xmlnode_put_attrib(error,"code","406");
        xmlnode_insert_cdata(error,"Invalid registration",-1);
        ehandler_send(ehandler_irc,err,jp->from->server);
        xmlnode_free(jp->x);
        return;
    }
    /* subscribe to the user */
    pth_mutex_acquire(&s__m,0,NULL);
    s=irc_find_default_session(jp->from);
    if(s!=NULL)
    {
        presence=xmlnode_new_tag("presence");
        xmlnode_put_attrib(presence,"to",jid_full(jp->from));
        if(s->pass!=NULL)
            spooler(sp,reg_host,"/registered?user=",s->nick,"&pass=",s->pass,"&server=",s->server,sp);
        else
            spooler(sp,reg_host,"/registered?user=",s->nick,"&pass=&server=",s->server,sp);
        xmlnode_put_attrib(presence,"from",spool_print(sp));
        xmlnode_put_attrib(presence,"type","unsubscribe");
        ehandler_send(ehandler_reg,presence,jp->from->server);
        --s->rc;
        s->default_session=0;
        if(s->rc==0)
            session_unlink(s);
    }    
    presence=xmlnode_new_tag("presence");
    xmlnode_put_attrib(presence,"to",jid_full(jp->from));
    spooler(sp,reg_host,"/registered?user=",nick,"&pass=",pass,"&server=",server,sp);
    xmlnode_put_attrib(presence,"from",spool_print(sp));
    xmlnode_put_attrib(presence,"type","subscribe");
    logon=xmlnode_dup(presence);
    ehandler_send(ehandler_irc,presence,jp->from->server);
    /* send back the result */
    jutil_iqresult(jp->x);
    jpacket_reset(jp);
    ehandler_send(ehandler_irc,xmlnode_dup(jp->x),jp->to->server);
    /* start the sesion */
    jutil_tofrom(logon);
    sp=spool_new(xmlnode_pool(logon));
    spooler(sp,"NULL%",server,"@",irc_host,"/",nick,sp);
    xmlnode_put_attrib(logon,"to",spool_print(sp));
    jp=jpacket_new(logon);
    if(s==NULL)
        irc_session_create(jp,1,pass);
    else
    {
        s->default_session=1;
        s->rc++;
    }
    pth_mutex_release(&s__m);
}

/*
takes a register packet from etherx, and decides what to do with it
*/
int irc_parse_reg_packet(jpacket jp)
/*
returns:
 0 on normal handle
 1 on error handle
input:
 jpacket jp: the packet from ehterx
*/
{
    session s;
    if(jp==NULL)return 1;
    switch(jp->type)
    {
    case JPACKET_IQ:
        switch(jpacket_subtype(jp))
        {
        case JPACKET__GET:
            if(NSCHECK(jp->iq,NS_REGISTER))
                irc_parse_iqget(jp);
            else
                xmlnode_free(jp->x);
            break;
        case JPACKET__SET:
            if(NSCHECK(jp->iq,NS_REGISTER))
                irc_parse_iqset(jp);
            else
                xmlnode_free(jp->x);
            break;
        default:
            xmlnode_free(jp->x);
        };
        break;
    case JPACKET_PRESENCE:
        switch(jpacket_subtype(jp))
        {
        case JPACKET__PROBE:
            s=irc_find_default_session(jp->from);
            if(s!=NULL)
            {
                spool sp=spool_new(jp->p);
                spooler(sp,"ISON :",jp->to->user,"\n",sp);
                tstream_write(s->stream,spool_print(sp),-1);
            }
            xmlnode_free(jp->x);
            break;
        case JPACKET__AVAILABLE:
            irc_auth_user(jp);
            break;
        case JPACKET__UNAVAILABLE:
             irc_parse_reg_unavailable(jp);
             break;
        default:
             xmlnode_free(jp->x);
        };
        break;
    case JPACKET_MESSAGE:
        irc_parse_reg_message(jp);
        break;
    case JPACKET_S10N:
        switch(jpacket_subtype(jp))
        {
        case JPACKET__SUBSCRIBE:
            subscribe_user(jp);
            break;
        case JPACKET__UNSUBSCRIBED:
            unsubscribed_user(jp);
            break;
        default:
            xmlnode_free(jp->x);
        };
        break;
    default:
        xmlnode_free(jp->x);
        break;
    };
    return 0;
}


