/*
IRC-Transport sessions.c: main session handler functions
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>

session sessions = NULL;

/*
called by the libxode framework when the pool dies (the tsream to IRC)
*/
void irc_session_cleanup(void *arg)
/*
returns:
 nothing
input:
 void *arg: the tstream that died
*/
{
    session s;
    if(arg==NULL)return;
    pth_mutex_acquire(&s__m,0,NULL);
    s=(session)arg;

    s->stream=NULL;
    irc_send_message(s,NULL,NULL,"groupchat","Disconnected",NULL);
    session_unlink(s);
    s->rc=0;
    pth_mutex_release(&s__m);
}

/*
called by the libxode framework when data arrives on the tstream
*/
void handle_read(thread t, void *buffer, int bufsz)
/*
returns:
 nothing
input:
 thread t: the tstream's thread
 void *buffer: the input buffer from the socket
 int bufsz: the size of the buffer
*/
{
    pool p;
    tstream ts;
    session s;
    log_debug(ZONE,"Recv Data: %s",buffer);
    if(t==NULL)return;
    ts=(tstream)t->data;
    if(ts==NULL) return;
    s=(session)ts->data;
    if(s==NULL) return;
    if(buffer==NULL) 
    { /* INIT */
        pool_cleanup(t->p,&irc_session_cleanup,(void*)s);
        return;
    }
    if(s->stream==NULL||s->rc==0) return;
    p=pool_new();
    parse_server(p,s,buffer);
    pool_free(p);
}

/*
called by the tstream framework when a socket error occurs
this function kills the tstream, and all refcounts to close the session
*/
void handle_err(thread t, tstream_buffer buffer)
/*
returns:
 nothing
input:
 thread t: the tstream's thread
 tstream_buffer buffer: the error buffer do-hickey from tstream
*/
{
    tstream ts;
    session s;
    if(t==NULL)return;
    ts=(tstream)t->data;
    if(ts==NULL) return;
    s=(session)ts->data;
    if(s==NULL) return;
    log_debug(ZONE,"Recv Error, closing session (%s)[%s]",jid_full(s->j),s->server);
    pth_mutex_acquire(&s__m,0,NULL);
    session_unlink(s);
    tstream_close(s->stream);
    pth_mutex_release(&s__m);
}

/*
removes a session from the sessions chain.
doesn't free anything.. the sesion is freed when the thread gets freed.
*/
void session_unlink(session s)
/*
returns:
 nothing
input:
 session s: the session to unlink
*/
{
    if(s==NULL) return;    
    pth_mutex_acquire(&s__m,0,NULL);
    if(sessions==s)
        sessions=s->next;
    if(s->next!=NULL)
        s->next->prev=s->prev;
    if(s->prev!=NULL)
        s->prev->next=s->next;    
    pth_mutex_release(&s__m);
    _debug_print_sessions();
}

/*
controls the main thread. as long as there are active locks, the session
remains running
*/
int irc_session_isdone(void *arg)
/*
returns:
 0 for normal operation
 1 if the thread should end
input:
 void *arg: the session to check
 */
{
    if(arg==NULL) return 1;
    if(((session)arg)->rc>0)
        return 0;
    else
        return 1;
}

/*
main session loop. all session threads start here
*/
void irc_session_main(void *arg)
/*
returns:
 nothing
input:
 void *arg: the thread for the session
*/
{
    thread t=(thread)arg;
    session s;
    spool sp,output;
    pool p;
    if(t==NULL)return;
    s=(session)(t->data);
        
    /* connect to irc server */
    log_debug(ZONE,"Connecting to server");
    s->stream=tstream_connect(s->port,s->server,&handle_read,&handle_err,(void*)s);
    /* always returns a valid stream, never returns NULL */
    output=spool_new(t->p);
    spooler(output,"USER jabber blah blah :",jid_full(s->j),"\nNICK ",s->nick,"\n",output);
    tstream_write(s->stream,spool_print(output),-1);
    log_debug(ZONE,"(%s) waiting for kill",jid_full(s->j));
    while(1)
    {
        pth_sleep(5);
        if(s->rc>0&&s->notify_list!=NULL)
        {
            char *u_list;
            users curr=s->notify_list;
            log_debug(ZONE,"Time event fired");
            p=pool_new();
            sp=spool_new(p);
            while(curr!=NULL)
            {
                spooler(sp,curr->user," ",sp);
                curr=curr->next;
            }
            u_list=spool_print(sp);
            sp=spool_new(p);
            spooler(sp,"ISON :",u_list,"\n",sp);
            tstream_write(s->stream,spool_print(sp),-1);
            pool_free(p);
        }
        else if(s->rc<=0)
            break;
    }
    pth_mutex_acquire(&s__m,0,NULL);
    session_unlink(s);
    pth_mutex_release(&s__m);
    log_warn(ZONE, "Closing down session for %s[%s]", jid_full(s->j),s->server);
    /* log user off IRC */
    if(s->stream!=NULL)
    {
        sp=spool_new(t->p);
        spooler(sp,"QUIT :",irc_get_quit_msg(),"\n",sp);
        tstream_write(s->stream,spool_print(sp),-1);
        pth_yield(NULL);
        tstream_close(s->stream);
    }
    while(s->notify_list!=NULL)
    {
        spool sp=spool_new(t->p);
        xmlnode presence=xmlnode_new_tag("presence");
        spooler(sp,s->notify_list->user,"@",reg_host,sp);
        xmlnode_put_attrib(presence,"from",spool_print(sp));
        xmlnode_put_attrib(presence,"to",jid_full(s->j));
        xmlnode_put_attrib(presence,"type","unavailable");
        ehandler_send(ehandler_reg,presence,s->j->server);
        s->notify_list=s->notify_list->next;
    }
    thread_free(t);
    return;
}
/*
creates a session, and sets the default flag
frees the jpacket param as well
*/
thread irc_session_create(jpacket jp,int default_conn,char *pass)
/*
returns:
 a pointer to the thread it creates
 NULL on failure
input:
 jpacket jp: available presence to start the sesion with
 int default_conn: the default session flag
*/
{
	thread t=NULL;
	session new=NULL;
    pth_attr_t attr;
    pool p=pool_new();
    if(jp==NULL)
    {   /* umm. supply more info please! */
        log_error(ZONE,"ASSERT FAILED:not enough info to start session");
        xmlnode_free(jp->x);
        return NULL;
    }
    pth_mutex_acquire(&s__m,0,NULL);
    t=thread_new();
    new = (session)pmalloc(t->p,sizeof(_session));
    new->t=t;
    new->stream=NULL;
	new->j=jid_new(t->p,jid_full(jp->from));
    new->nick=pstrdup(t->p,jp->to->resource);
    new->server=irc_get_server(t->p,jid_full(jp->to));
    new->port=atoi(irc_get_port(jp->p,jid_full(jp->to)));
    if(!default_conn)
    {
        new->channel_list=pmalloc(p,sizeof(_chans));
        new->channel_list->p=p;
        new->channel_list->next=NULL;
        new->channel_list->id=pstrdup(p,jp->to->user);
        new->channel_list->user_list=NULL;
        new->channel_list->channel=irc_get_channel(p,jid_full(jp->to));
    }
    else
    {
        new->channel_list=NULL;
        log_debug(ZONE,"Starting a default thread");
    }
    new->pass=pass;
    new->rc=1;
    new->default_session=default_conn;
    new->notify_list=NULL;
    new->prev=NULL;
    new->next = sessions;
    if(sessions!=NULL)
        sessions->prev=new;
    sessions = new;
    t->data=(void*)new;
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    if(!default_conn)
        irc_send_message(new,pstrdup(t->p,new->channel_list->channel),NULL,"groupchat","Please Wait while IRC-Transport connects to the server and/or joins the channel...",NULL);
    t->id = pth_spawn(attr, (void *)&irc_session_main, (void *)t);
    pth_attr_destroy(attr);
    xmlnode_free(jp->x); /*new session free*/
    pth_mutex_release(&s__m);
	return t;
}

/*
parts the user from a channel, reduces the ref count by one
*/
void irc_session_end(session s,char *channel,char *reason)
/*
returns:
 nothing
input:
 session s: the session this is for
 char *channel: the channel we want to quit from.
*/
{
    spool sp;
    chans prev,curr;
    if(s==NULL) return;
    pth_mutex_acquire(&s__m,0,NULL);
    if(channel==NULL)
    { /* part from ALL channels */
        prev=NULL;     
        for(;s->channel_list!=NULL;s->channel_list=s->channel_list->next)
        {
             irc_send_presence(s,s->channel_list->channel,s->nick,JPACKET__UNAVAILABLE,NULL);
             sp=spool_new(s->channel_list->p);
             if(reason==NULL)
                 spooler(sp,"PART ",irc_get_channel(s->channel_list->p,s->channel_list->id)," :",irc_get_part_msg(),"\n",sp);
             else
                 spooler(sp,"PART ",irc_get_channel(s->channel_list->p,s->channel_list->id)," :",reason,"\n",sp);    
             tstream_write(s->stream,spool_print(sp),-1);
             if(prev!=NULL)
                 pool_free(prev->p); /*free working pool*/
             prev=s->channel_list;
             s->rc--;
        }
        if(prev!=NULL)
            pool_free(prev->p); /*free working pool*/
        pth_mutex_release(&s__m);
        return;
    }
    for(prev=NULL,curr=s->channel_list;curr!=NULL;prev=curr,curr=curr->next)
    {
        if(irc_channel_cmp(curr->channel,channel)==0)
        {
            irc_send_presence(s,curr->channel,s->nick,JPACKET__UNAVAILABLE,NULL);
            if(prev==NULL)
                s->channel_list=curr->next;
            else
                prev->next=curr->next;
            sp=spool_new(curr->p);
            if(reason==NULL)
                spooler(sp,"PART ",channel," :",irc_get_part_msg(),"\n",sp);
            else
                spooler(sp,"PART ",channel," :",reason,"\n",sp);    
            tstream_write(s->stream,spool_print(sp),-1);
            pool_free(curr->p); /*free working pool*/
            s->rc--;
            if(s->rc==0)
            {
                session_unlink(s);    
                tstream_close(s->stream);    
            }
            break;
        }
    }  
    pth_mutex_release(&s__m);
}

/*
finds a session, given a jid and a server
NOTE that a session is based on jid+server.. a single user may
have more than one active session, on different servers
*/
session irc_session_find_by_jid(jid j, char *server)
/*
returns:
 NULL on failure to find the session
 a pointer to the session found
input:
 jid j: the jid to search for
 char *server: the name of the server to check against
*/
{
    session next;
    if(j==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    log_debug(ZONE, "[session] finding session for (%s)[%s]", jid_full(j),server );
    if( sessions == NULL ) return NULL;
    for(next = sessions; next != NULL; next = next->next)
        if(jid_cmp(j, next->j) == 0)
            if(server==NULL||strcasecmp(next->server,server)==0)
            {
                pth_mutex_release(&s__m);
                return next;
            }
    pth_mutex_release(&s__m);
    return NULL;
}

