/*
IRC-Transport util.c: general utility functions for IRC-transport 
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include "irctrans.h"

extern session sessions;

/*
prints out a list of sessions
*/
void _debug_print_sessions(void)
/*
returns:
 nothing
input:
 none
*/
{
    session s;
    pth_mutex_acquire(&s__m,0,NULL);
    log_debug("DEBUG","LISTING SESSIONS:");
    for(s=sessions;s!=NULL;s=s->next)
        log_debug("DEBUG","%s (%d Active Locks), default: %d",jid_full(s->j),s->rc,s->default_session);
    pth_mutex_release(&s__m);
}


/*
takes a jid, and takes off the resource bits
*/
void irc_strip_jid(jid j)
/*
returns:
 nothing
input:
 jid j: the jid to strip
*/
{
    if(j==NULL) return;
    jid_set(j,NULL,JID_RESOURCE);
}

/*
returns the session id for a certain channel
an id is the user part that the user sent,
so that the client can route messages to the right windows..
since #foo@irc.j.o is the same as foo@irc.j.o or foo%irc.linux.com@j.o or... (etc etc..)
*/
char *irc_get_id(session s,char *channel)
/*
returns:
 a pointer to the id,
 NULL if the channel wasn't found
input:
 session s: the session to look for the channel
 char *channel: the name of the channel to find the id for
*/
{
    chans curr;
    pth_mutex_acquire(&s__m,0,NULL);
    if(s==NULL||channel==NULL) return NULL;
    curr=irc_find_channel_by_name(s,channel);
    pth_mutex_release(&s__m);
    if(curr!=NULL) return curr->id;
    return NULL;
}

/*
deletes a user from the channel user list
*/
void irc_del_user(session s,char *chan,char *nick)
/*
returns:
 nothing
input:
 session s: the session to delete from
 char *chan: the name of the channel to delete
 char *nick: the nick of the user to remove
*/
{
    irc_del_user_by_id(s,irc_get_id(s,chan),nick);
}

/*
deletes a user from the channel user list
*/
void irc_del_user_by_id(session s,char *id,char *nick)
/*
returns:
 nothing
input:
 session s: the session to delete from
 char *id: the id of the channel to delete
 char *nick: the nick of the user to remove
*/
{
    chans curr;
    users u_list,prev=NULL;
    if(s==NULL||id==NULL||nick==NULL) return;
    pth_mutex_acquire(&s__m,0,NULL);
    curr=irc_find_channel_by_id(s,id);
    if(curr==NULL) return;
    for(u_list=curr->user_list;u_list!=NULL;prev=u_list,u_list=u_list->next)
    {
        if(irc_nick_cmp(nick,u_list->user)==0)
        {
            if(prev==NULL)
            {
                curr->user_list=u_list->next;
                break;
            }
            else
            {
                prev->next=u_list->next;
                break;
            }
        }
    }
    pth_mutex_release(&s__m);
}

/*
adds a user to the ISON notify list
*/
void irc_add_notify(session s,char *nick)
/*
returns:
 nothing
input:
 session s: the session to add this to (default session)
 char *nick: the nick to monitor for presence
*/
{
    users new;
    pth_mutex_acquire(&s__m,0,NULL);
    if(s==NULL||nick==NULL)return;
    /* check to see if it already is in the DB */
    for(new=s->notify_list;new!=NULL;new=new->next)
        if(irc_nick_cmp(new->user,nick)==0)
        {
            pth_mutex_release(&s__m);
            return; /* it's already in here! */
        }
    new=pmalloc(s->t->p,sizeof(_users));
    new->user=pstrdup(s->t->p,nick);
    new->voice=0;
    new->ops=0;
    new->next=s->notify_list;
    s->notify_list=new;
    pth_mutex_release(&s__m);
}

/*
deletes a user from the ISON notify list
*/
void irc_del_notify(session s,char *nick)
/*
returns:
 nothing
input:
 session s: the session to delete from
 char *nick: the nickname to remove from the notify list
*/
{
    users curr,prev=NULL;
    if(s==NULL||nick==NULL)return;
    pth_mutex_acquire(&s__m,0,NULL);
    for(curr=s->notify_list;curr!=NULL;prev=curr,curr=curr->next)
    {
        if(j_strcmp(curr->user,nick)==0)
        {
            if(prev==NULL)
                s->notify_list=curr->next;
            else
                prev->next=curr->next;
            break;
        }
    }
    pth_mutex_release(&s__m);
}

/*
sends out the nesseccary presnce to the user
->ops represents the current setting (on/off)
->voice representes the new value (on/off)
if voice is off and ops is on, then the user went offline
if voice is on and ops is off, then the user went online
*/
void irc_update_notify_pres(session s)
/*
returns:
 nothing
input:
 session s: the session to check against.
*/
{
    users curr;
    if(s==NULL)return;
    pth_mutex_acquire(&s__m,0,NULL);
    curr=s->notify_list;
    while(curr!=NULL)
    {
        if(curr->ops&&!curr->voice)
        {
            /* went offline */
            xmlnode presence=xmlnode_new_tag("presence");
            spool sp=spool_new(xmlnode_pool(presence));
            spooler(sp,curr->user,"@",reg_host,sp);
            xmlnode_put_attrib(presence,"to",jid_full(s->j));
            xmlnode_put_attrib(presence,"from",spool_print(sp));
            xmlnode_put_attrib(presence,"type","unavailable");
            xmlnode_insert_cdata(xmlnode_insert_tag(presence,"status"),"Nick not found",-1);
            ehandler_send(ehandler_irc,presence,s->j->server);
            curr->ops=curr->voice;
        }
        else if(!curr->ops&&curr->voice)
        {
            /* now online */
            xmlnode presence=xmlnode_new_tag("presence");
            spool sp=spool_new(xmlnode_pool(presence));
            spooler(sp,curr->user,"@",reg_host,sp);
            xmlnode_put_attrib(presence,"to",jid_full(s->j));
            xmlnode_put_attrib(presence,"from",spool_print(sp));
            xmlnode_insert_cdata(xmlnode_insert_tag(presence,"status"),"Nick now Online",-1);
            ehandler_send(ehandler_irc,presence,s->j->server);
            curr->ops=curr->voice;
        }
        curr=curr->next;
    }
    pth_mutex_release(&s__m);
}

/*
add a user to the channel user list
*/
users irc_add_user(session s,char *chan, char *nick)
/*
returns:
 the user added
 NULL if there was an error
input:
 session s: the session to add to
 char *chan: the channel to add the user to.
 char *nick: the nickname to add to the channel
*/
{
    return irc_add_user_by_id(s,irc_get_id(s,chan),nick);
}

/*
add a user to the channel user list
*/
users irc_add_user_by_id(session s,char *id,char *nick)
/*
returns:
 the user added
 NULL if there was an error
input:
 session s: the session to add to
 char *id: the channel id to add the user to.
 char *nick: the nickname to add to the channel
*/
{
    chans curr;
    users u_list,prev=NULL;
    if(s==NULL||id==NULL||nick==NULL) return NULL;
    curr=irc_find_channel_by_id(s,id);
    if(curr==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    u_list=curr->user_list;
    /* find the last user, make sure that this user
       doesn't already exist. */
    while(u_list!=NULL)
    {
        if(irc_nick_cmp(nick,u_list->user)==0)
        {
            /* this user already exists */
            pth_mutex_release(&s__m);
            return u_list;
        }
        prev=u_list;
        u_list=u_list->next;
    }
    u_list=pmalloc(curr->p,sizeof(_users));
    u_list->ops=u_list->voice=0;
    if(nick[0]=='@')
    {
        nick++;
        u_list->ops=1;
    }
    else if(nick[0]=='+')
    {
        nick++;
        u_list->voice=1;
    }
    u_list->user=pstrdup(curr->p,nick);
    u_list->next=curr->user_list;
    curr->user_list=u_list;
    pth_mutex_release(&s__m);
    return u_list;
}

/*
finds a user in the channel list
*/
users irc_get_user(session s,char *channel, char *nick)
/*
returns:
 the user found
 NULL if no user was found
input:
 session s: the session to look in
 char *channel: the channel to find the user in
 char *nick: the nick we are looking for
*/
{
    chans curr;
    users u_list;
    if(s==NULL||channel==NULL||nick==NULL) return NULL;
    curr=irc_find_channel_by_name(s,channel);
    if(curr==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    u_list=curr->user_list;
    while(u_list!=NULL)
    {
        if(irc_nick_cmp(nick,u_list->user)==0)
        {
            pth_mutex_release(&s__m);
            return u_list;
        }
        u_list=u_list->next;
    }
    pth_mutex_release(&s__m);
    return NULL;
}

/*
sets a mode for the user (ops and voice)
*/
void irc_set_mode(users u,int ops,int voice)
/*
returns;
 nothing
input:
 users u: the user to update
 int ops: the new ops vaule
 int voice: the new voice value
*/
{
    if(u==NULL) return;
    u->ops=ops;
    u->voice=voice;
}

/*
compares two nicknames, ignoreing ops and voice flags.
*/
int irc_nick_cmp(char *nick1,char *nick2)
/*
returns:
 -1 if either are NULL
 return value from strcasecmp for the two strings
input:
 char *nick1: duh
 char *nick2: duh
*/
{
    char *n1=nick1,*n2=nick2;
    if(n1==NULL||n2==NULL) return -1;
    if(n1[0]=='@'||n1[0]=='+') n1++;
    if(n2[0]=='@'||n2[0]=='+') n2++;
    return strcasecmp(n1,n2);
}

/* 
finds the FIRST id for a given nick, based on session.
if a user is part of more than one channel, this will only return the first one..
this makes it wierd when replying to chat, as you may get a reply from a different
user than the one you sent to, but heck, it still gets there doesn't it? quit your bitchin'
*/        
char *irc_get_id_by_user(session s,char *nick)
/*
returns:
 a pointer to the id the user is in
input:
 session s: the session to look in
 char *nick: the user you are looking for
*/
{
    chans curr;
    if(s==NULL||nick==NULL) return NULL;
    curr=s->channel_list;
    pth_mutex_acquire(&s__m,0,NULL);
    while(curr!=NULL)
    {
        users u_list=curr->user_list;
        while(u_list!=NULL)
        {
            if(irc_nick_cmp(nick,u_list->user)==0)
            {
                pth_mutex_release(&s__m);
                return curr->id;
            }
            u_list=u_list->next;
        }
        curr=curr->next;
    } 
    pth_mutex_release(&s__m);
    return NULL;
}

/*
strips out those nasty color codes
overwrites input with the new data
*/
void irc_strip_mirc_codes(char *input)
/*
returns:
 nothing
input:
 char *input: the string to convert, this gets overwritten iwth the result *MUST NOT BE READ ONLY* 
*/
{
    char *curr=input,*write=input;
    int a,w;
    long c;
    if(input==NULL) return;
    for(a=0,c=0,w=0;a<strlen(input);a++)
    {
        c=(long)curr[a];
        /*(  bold||underline||reverse||  plain)*/
        if(c==0x2||c==0x1F  ||c==0x16||c==0x0F)
        {
            /* skip this char */
            continue;
        }
        else if(c==0x3) /* one of these codes: # ## #,# #,## ##,# ##,## */
        { 
            c=curr[++a];
            if(isdigit(c))
            { /* # ## #,# #,## ##,# or ##,## */
                c=curr[++a];
                if(isdigit(c))
                { /* ## ##,# ##,## */
                    if(c==',')
                    { /* ##,# ##,## */
                        c=curr[a+2];
                        if(isdigit(c)) c=curr[++a];
                    }
                }
                else if(c==',')
                { /* #,# or #,## */    
                    c=curr[a+2];
                    if(isdigit(c)) c=curr[++a];
                }
            }
        }
        /* write the next char */
        write[w++]=(char)c;
    }
    write[w]='\0';
}

/* 
function to match '#Jabber' with 'jabber'
or '.jabber' with '&jabber'
NULL != NULL
*/
int irc_channel_cmp(const char *chan1,const char *chan2)
/*
returns:
 -1 if either are NULL
 the return value from strcasecmp
input:
 const char *char1,const char *char2: the values to compare
*/
{
    pool p;
    int ret=-1;
    char *c1,*c2;
    if(chan1==NULL||chan2==NULL) return ret;
    p=pool_new();
    c1=pstrdup(p,chan1);
    c2=pstrdup(p,chan2);
    /* normalize all channel names
        #foo => foo
        foo  => foo
        &foo => .foo
        .foo => .foo
    */
    if(c1[0]=='#') c1++;
    if(c1[0]=='&') c1[0]='.';
    if(c2[0]=='#') c2++;
    if(c2[0]=='&') c2[0]='.';
    if(c1!=NULL&&c2!=NULL)
        ret=strcasecmp(c1,c2);
    pool_free(p);
    return ret;
}

/*
extracts the channel name from a jid string
*/
char *irc_get_channel(pool p, char *input)
/*
returns:
 a pool'd copy o fthe channel using the supplied pool
 NULL if it could not find the channel
input:
 pool p: a pool to use to copy the data
 char *input: the jid string to extract from
*/
{
    char *curr,*next;
    char *ret=pstrdup(p,input);
    spool s;
    if(p==NULL||input==NULL) return NULL;
    curr=ret;
    next=strchr(curr,'%');
    if(next==NULL) next=strchr(curr,'@');
    if(next==NULL) return NULL;
    next[0]='\0';
    s=spool_new(p);
    if(curr[0]=='.') curr[0]='&';
    if(curr[0]!='#'&&curr[0]!='&') 
        spooler(s,"#",curr,s);
    else
        spooler(s,curr,s);
    return spool_print(s);
}

/*
extracts the server name from a jid string
*/
char *irc_get_server(pool p,char *input)
/*
returns:
 a pool'd copy of the server name using the supplied pool
 NULL if it could not find the server name
input:
 pool p: a pool to use to copy the data
 char *input: the jid string to extract from
*/
{
    char *curr,*next;
    char *ret=pstrdup(p,input);
    if(p==NULL||input==NULL) return NULL;
    curr=ret;
    next=strchr(curr,'%');
    if(next==NULL) 
        return pstrdup(p,default_server);
    curr=next+1;
    ret=curr;
    next=strchr(curr,'%');
    if(next==NULL) next=strchr(curr,'@');
    if(next==NULL) 
        return pstrdup(p,default_server);
    next[0]='\0';
    return ret;
}

/*
extracts the port name from a jid string
*/
char *irc_get_port(pool p,char *input)
/*
returns:
 a pool'd copy of the port string using the supplied pool
 NULL if it could not find the port
input:
 pool p: a pool to use to copy the data
 char *input: the jid string to extract from
*/
{
    char *curr,*next;
    char *ret=pstrdup(p,input);
    if(p==NULL||input==NULL) return NULL;
    curr=ret;
    next=strchr(curr,'%');
    if(next==NULL) 
        return pstrdup(p,default_port);    
    next=strchr(curr,'%');
    if(next==NULL)
        return pstrdup(p,default_port);    
    curr=next+1;
    next=strchr(curr,'@');
    if(next==NULL);
        return pstrdup(p,default_port);    
    ret=curr;
    next[0]='\0';
    return ret;
}

/*
finds a channel in a session, given a channel name
*/
chans irc_find_channel_by_name(session s,char *channel)
/*
returns:
 a pointer to the channel found
 NULL if no channel was found
input:
 session s: a session to search through
 char *channel: the channel name to find
*/
{
    chans curr;
    if(s==NULL||channel==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    for(curr=s->channel_list;curr!=NULL;curr=curr->next)
    {
        if(irc_channel_cmp(channel,curr->channel)==0)
        {
            pth_mutex_release(&s__m);
            return curr;
        }
    }
    pth_mutex_release(&s__m);
    return NULL;
}

/*
finds a channel, given a channel id string
*/
chans irc_find_channel_by_id(session s,char *id)
/*
returns:
 a pointer to the channel found
 NULL if no channel was found
input:
 session s: the session to search through
 char *id: the id string to look for
*/
{
    chans curr;
    if(s==NULL||id==NULL) return NULL;
    pth_mutex_acquire(&s__m,0,NULL);
    for(curr=s->channel_list;curr!=NULL;curr=curr->next)
    {
        if(strcasecmp(id,curr->id)==0)
        {
            pth_mutex_release(&s__m);    
            return curr;
        }
    } 
    pth_mutex_release(&s__m);
    return NULL;
}

/*
sends an error message to the jabber client
*/
void irc_send_error(jid to,char *from, char *code, char *msg)
/*
returns:
 nothing
input:
 jid to: the jid to send to (usually always s->j)
 char *from: the jid to say it's from
 char *code: the error code to include in the msg
 char *msg: the text of the error message
*/
{
    xmlnode error,err;
    if(to==NULL||from==NULL||code==NULL||msg==NULL) return;
    error=xmlnode_new_tag("message");
    xmlnode_put_attrib(error,"to",jid_full(to));
    xmlnode_put_attrib(error,"from",from);
    xmlnode_put_attrib(error,"type","error");
    err=xmlnode_insert_tag(error,"error");
    xmlnode_put_attrib(err,"code",code);
    xmlnode_insert_cdata(err,msg,strlen(msg));
    ehandler_send(ehandler_irc,error,to->server);
}

/*
sends a message to a jabber user
*/
void irc_send_message(session s,char *channel,char *nick,char *type,char *bodyIn, char *subject)
/*
returns:
 nothing
input:
 session s: a session to send to (uses s->j to send the message to)
 char *channel: the channel name it's from (to build the from jid) NULL if all channels
 char *nick: the nick to say it's from( to build the jid) NULL if from the server
 char *type: the type ofo the message, NULL is 'normal' or specify 'groupchat' or 'chat', etc
 char *bodyIn: the body of th emessage
 char *subject: the subject of the message
*/
{
    xmlnode message;
    char *body;
    chans curr;
    spool sp;
    if(s==NULL||s->j==NULL||bodyIn==NULL) return;
    pth_mutex_acquire(&s__m,0,NULL);
    if(channel!=NULL&&channel[0]=='&') channel[0]='.';
    curr=irc_find_channel_by_name(s,channel);
    message=xmlnode_new_tag("message");
    body=pstrdup(xmlnode_pool(message),bodyIn);
    sp=spool_new(xmlnode_pool(message));
    if(channel==NULL&&nick==NULL&&s->default_session==1&&s->rc==1)
    { /* *phew* basically, don't say 'nuthin unless we're on a channel */
        pth_mutex_release(&s__m);
        return;
    }    
    else if((channel==NULL)||(curr==NULL))
        spooler(sp,s->server,"@",irc_host,sp);
    else if(nick==NULL)
        spooler(sp,curr->id,"@",irc_host,sp);
    else
        spooler(sp,curr->id,"@",irc_host,"/",nick,sp);
    xmlnode_put_attrib(message,"to",jid_full(s->j));
    xmlnode_put_attrib(message,"from",spool_print(sp));
    if(channel==NULL)
        xmlnode_put_attrib(message,"type","chat");    
    else if(type!=NULL)
        xmlnode_put_attrib(message,"type",type);
    if(subject!=NULL)
    {
        irc_strip_mirc_codes(subject);
        subject=str_to_UTF8(message->p,subject);
        xmlnode_insert_cdata(xmlnode_insert_tag(message,"subject"),subject,-1);
    }
    irc_strip_mirc_codes(body);
    body=str_to_UTF8(message->p,body);
    xmlnode_insert_cdata(xmlnode_insert_tag(message,"body"),body,-1);
    ehandler_send(ehandler_irc,message,s->j->server);
    pth_mutex_release(&s__m);
}
/*
sends presence to a jabber user
*/
void irc_send_presence(session s,char *channel,char *nick,int type,char *status)
/*
returns:
 nothing
input:
 session s: the session to send to.. sends presence to s->j
 char *channel: the channel to say it's from to build the from JID should not be NULL
 char *nick: the nick to send presence from,.. should not be null
 int type: the JPACKET__* type to send a presence type
 char *status: the status message
*/
{
    users u=irc_get_user(s,channel,nick);
    xmlnode presence=jutil_presnew(type,jid_full(s->j),status);
    spool sp=spool_new(presence->p);
    if(channel!=NULL&&channel[0]=='&') channel[0]='.';
    if(u!=NULL&&u->ops)
        spooler(sp,irc_get_id(s,channel),"@",irc_host,"/@",u->user,sp);
    else if(u!=NULL&&u->voice)
        spooler(sp,irc_get_id(s,channel),"@",irc_host,"/+",u->user,sp);
    else if(u!=NULL)
        spooler(sp,irc_get_id(s,channel),"@",irc_host,"/",u->user,sp);
    else
        spooler(sp,irc_get_id(s,channel),"@",irc_host,"/",s->nick,sp);
    xmlnode_put_attrib(presence,"from",spool_print(sp));
    ehandler_send(ehandler_irc,presence,s->j->server);
}

/* send an outgoing line to the connection */
void svc_irc_line(session s, char *cmd, char *data, int argc, ...)
{
    static char buff[512], args[512];
    char *str;
    int strsz, i, argsz, flag;
    va_list ap;

    va_start(ap, argc);
    argsz = 0;
    args[argsz] = '\0';
    for(i = 0; i < argc; i++)
    {
        str = va_arg(ap,char *);
        if(str == NULL) continue;
        if(strlen(str) + argsz >= 500) continue;
        strcat(args,str);
        argsz += strlen(str);
        strcat(args," ");
        argsz++;
    }
    va_end(ap);

    if(data == NULL)
    {
        strsz = snprintf(buff,512,"%s %s\r\n",cmd,args);
        log_debug("IRC-OUT1",buff);
        tstream_write(s->stream,buff,strsz);
        return;
    }

    argsz = strlen(data);
    flag = 0;
    for(i = 1; i < argsz; i++)
    {
        if(data[i] == '\n')
        {
            if(data[i-1] == '\r')
                data[i-1] = '\0';
            data[i] = '\0';
            flag = 1;
            strsz = snprintf(buff,512,"%s %s:%s\r\n",cmd,args,data);
        }else if(i == 200){
            strsz = snprintf(buff,512,"%s %s:%.200s\r\n",cmd,args,data);
            flag = 1;
        }


        if(flag)
        {
            log_debug("IRC-OUT2",buff);
            tstream_write(s->stream,buff,strsz);
            flag = 0;
            data = data + i + 1;
            i = -1; /* get incremented by for */
            argsz = strlen(data);
        }
    }

    if(strlen(data) > 0)
    {
        strsz = snprintf(buff,512,"%s %s:%s\r\n",cmd,args,data);
        log_debug("IRC-OUT3",buff);
        tstream_write(s->stream,buff,strsz);
    }
}

/*
returns the next PART message
*/
char *irc_get_part_msg(void)
{
    static char *default_text="Jabber (www.jabbercentral.com) IRC Transport Logging off channel";
    if(part_list==NULL)
        return default_text;    
    part_list=part_list->next;    
    return part_list->user;
}

/*
returns the next QUIT message
*/
char *irc_get_quit_msg(void)
{
    static char *default_text="Jabber (www.jabbercentral.com) IRC Transport Logging off channel";
    if(part_list==NULL)
        return default_text;    
    quit_list=quit_list->next;    
    return quit_list->user;
}
